import json
import logging
import idna
import requests
from certbot import errors
from certbot.plugins import dns_common

try:
    import certbot.compat.os as os
except ImportError:
    import os

logger = logging.getLogger(__name__)


class Authenticator(dns_common.DNSAuthenticator):
    """DNS Authenticator for Buddy

    This plugin enables usage of Buddy rest API to complete ``dns-01`` challenges."""

    description = "Automates dns-01 challenges using Buddy internal API"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.host = ""
        self.token = ""
        self.machine = ""
        self.workspace_id = ""
        self.tunnel_id = ""
        self.domain = ""

    @classmethod
    def add_parser_arguments(cls, add, **kwargs):
        super(Authenticator, cls).add_parser_arguments(
            add, default_propagation_seconds=120
        )

    def more_info(self):
        return self.description

    def _setup_credentials(self):
        host = os.getenv("BUDDY_HOST")
        token = os.getenv("BUDDY_TOKEN")
        machine = os.getenv("BUDDY_MACHINE")
        workspace_id = os.getenv("BUDDY_WORKSPACE_ID")
        tunnel_id = os.getenv("BUDDY_TUNNEL_ID")
        domain = os.getenv("BUDDY_DOMAIN")
        if host is None:
            raise errors.PluginError("host not defined")
        if token is None:
            raise errors.PluginError("token not defined")
        if machine is None:
            raise errors.PluginError("machine not defined")
        if workspace_id is None:
            raise errors.PluginError("workspace_id not defined")
        if tunnel_id is None:
            raise errors.PluginError("tunnel_id not defined")
        if domain is None:
            raise errors.PluginError("domain not defined")
        self.host = host
        self.token = token
        self.machine = machine
        self.workspace_id = workspace_id
        self.tunnel_id = tunnel_id
        self.domain = domain

    def _perform(self, domain, validation_name, validation):
        decoded_domain = idna.decode(domain)
        try:
            self._api_client().add_txt_record(decoded_domain, validation_name, validation)
        except ValueError as err:
            raise errors.PluginError("Cannot add txt record: {err}".format(err=err))

    def _cleanup(self, domain, validation_name, validation):
        decoded_domain = idna.decode(domain)
        try:
            self._api_client().del_txt_record(decoded_domain, validation_name, validation)
        except ValueError as err:
            raise errors.PluginError("Cannot remove txt record: {err}".format(err=err))

    def _api_client(self):
        return _ApiClient(self.host, self.token, self.machine, self.workspace_id, self.tunnel_id, self.domain)


class _ApiClient:
    def __init__(self, host, token, machine, workspace_id, tunnel_id, domain):
        """Initialize class managing a domain within Buddy API

        :param str host: API host
        :param str token: token
        :param str machine: machine
        :param str workspace_id: workspace ID
        :param str tunnel_id: tunnel ID
        :param str domain: domain
        """
        self.host = host
        self.token = token
        self.machine = machine
        self.workspace_id = workspace_id
        self.tunnel_id = tunnel_id
        self.domain = domain
        self.session = requests.Session()
        self.session.headers.update({
            "Accept": "application/json",
            "Content-Type": "application/json",
        })

    def _post_request(self, url, payload):
        """Perform a POST request to Buddy API
        :param url: relative URL
        :param payload: request body"""
        url = self.host + url
        logger.debug("POST %s", url)
        with self.session.post(url, json=payload, verify=False) as res:
            try:
                result = res.json()
            except json.decoder.JSONDecodeError:
                raise errors.PluginError("no JSON in API response")
            if res.status_code == requests.codes.ok:
                return result
            if result["errors"]:
                raise errors.PluginError(result["errors"][0]["message"])
            raise errors.PluginError("something went wrong")

    def add_txt_record(self, domain, name, value, ttl=300):
        """Add a TXT record to a domain
        :param str domain: name of domain to lookup
        :param str name: record key in zone
        :param str value: value of record
        :param int ttl: optional ttl of record"""
        if domain != self.domain:
            raise errors.PluginError("wrong domain")
        # foo.bar _acme-challenge.foo.bar MIFwJwRzJuLEknEfCY2PQwhzE-yf1WIisPFjCWAlKEs
        subdomain = name.replace("." + domain, "")
        self._post_request("/tunnel/dns/create", {
            "token": self.token,
            "machine": self.machine,
            "workspaceId": self.workspace_id,
            "tunnelId": self.tunnel_id,
            "domain": self.domain,
            "subdomain": subdomain,
            "type": "TXT",
            "ttl": ttl,
            "value": value
        })

    def del_txt_record(self, domain, name, value):
        """Delete a TXT record from a domain
        :param str domain: name of domain to lookup
        :param str name: record key in zone
        :param str value: value of record"""
        if domain != self.domain:
            raise errors.PluginError("wrong domain")
        subdomain = name.replace("." + domain, "")
        self._post_request("/tunnel/dns/remove", {
            "token": self.token,
            "machine": self.machine,
            "workspaceId": self.workspace_id,
            "tunnelId": self.tunnel_id,
            "domain": self.domain,
            "subdomain": subdomain,
            "type": "TXT",
        })



"""Tests for certbot_dns_buddy.dns_buddy."""

import io
import logging
import sys
from unittest import mock

try:
    import certbot.compat.os as os
except ImportError:
    import os
from certbot.plugins import dns_test_common
from certbot.plugins.dns_test_common import DOMAIN
from certbot.tests import util as test_util

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

FAKE_TOKEN = "XXXX"
FAKE_HOST = "https://foo.bar"
FAKE_MACHINE = "abc"
FAKE_WORKSPACE_ID = "123"
FAKE_TUNNEL_ID = "456"
FAKE_DOMAIN = "foo.bar"

class AuthenticatorTest(test_util.TempDirTestCase, dns_test_common.BaseAuthenticatorTest):
    """Class to test the Authenticator."""
    def setUp(self):
        super(AuthenticatorTest, self).setUp()

        self.config = mock.MagicMock()

        os.environ["BUDDY_TOKEN"] = FAKE_TOKEN
        os.environ["BUDDY_HOST"] = FAKE_HOST
        os.environ["BUDDY_MACHINE"] = FAKE_MACHINE
        os.environ["BUDDY_WORKSPACE_ID"] = FAKE_WORKSPACE_ID
        os.environ["BUDDY_TUNNEL_ID"] = FAKE_TUNNEL_ID
        os.environ["BUDDY_DOMAIN"] = FAKE_DOMAIN

        from certbot_dns_buddy_internal.dns_internal import Authenticator
        self.auth = Authenticator(self.config, "buddy")
        self.mock_client = mock.MagicMock(default_propagation_seconds=15)
        self.auth._api_client = mock.MagicMock(return_value=self.mock_client)

        try:
            from certbot.display.util import notify
            notify_patch = mock.patch('certbot._internal.main.display_util.notify')
            self.mock_notify = notify_patch.start()
            self.addCleanup(notify_patch.stop)
            self.old_stdout = sys.stdout
            sys.stdout = io.StringIO()
        except ImportError:
            self.old_stdout = sys.stdout

    def tearDown(self):
        sys.stdout = self.old_stdout

    def test_perform(self):
        """Tests the perform function to see if client method is called"""
        self.auth.perform([self.achall])
        expected = [
            mock.call.add_txt_record(DOMAIN, "_acme-challenge." + DOMAIN, mock.ANY)
        ]
        self.assertEqual(expected, self.mock_client.mock_calls)

    def test_cleanup(self):
        """Tests cleanup method to see if client method is called"""
        self.auth._attempt_cleanup = True
        self.auth.cleanup([self.achall])
        expected = [
            mock.call.del_txt_record(DOMAIN, "_acme-challenge." + DOMAIN, mock.ANY)
        ]
        self.assertEqual(expected, self.mock_client.mock_calls)


from setuptools import setup
from setuptools import find_packages
from os import path

version = "0.0.1"

install_requires = [
    "certbot>=0.31.0",
    "setuptools",
    "requests",
    "requests-mock",
    "idna",
]

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.rst")) as f:
    long_description = f.read()

setup(
    name="certbot-dns-buddy-internal",
    version=version,
    description="Buddy DNS internal Authenticator plugin for Certbot",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/buddy/certbot-dns-internal",
    author="Michal Hankus",
    license="Apache License 2.0",
    python_requires=">=3.5.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Plugins",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Security",
        "Topic :: System :: Installation/Setup",
        "Topic :: System :: Networking",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    entry_points={
        "certbot.plugins": [
            "dns-buddy-internal = certbot_dns_buddy_internal.dns_internal:Authenticator"
        ]
    },
    test_suite="certbot_dns_buddy_internal",
)
